import click

from alpa.cli.alpa_repo import create, delete, request_package
from alpa.cli.local_repo import (
    show_history,
    switch,
    commit,
    pull,
    push,
    list_,
    genspec,
    add,
    get_pkg_archive,
    mockbuild,
    create_packit_config,
)
from alpa.repository import AlpaRepo

# TODO: get rid of the repetitive _Repo(Path(getcwd()))


@click.group()
def entry_point() -> None:
    pass


# commands that don't require git repo at all


@entry_point.command("clone")
@click.argument("url", type=str)
@click.option("--fork", is_flag=True, default=False)
def clone(url: str, fork: bool) -> None:
    """Clone and prepare Alpa repository"""
    AlpaRepo.clone(url, fork)


entry_point.add_command(create)
entry_point.add_command(delete)
entry_point.add_command(request_package)

entry_point.add_command(show_history)
entry_point.add_command(switch)
entry_point.add_command(commit)
entry_point.add_command(pull)
entry_point.add_command(push)
entry_point.add_command(list_)
entry_point.add_command(genspec)
entry_point.add_command(add)
entry_point.add_command(get_pkg_archive)
entry_point.add_command(mockbuild)
entry_point.add_command(create_packit_config)


if __name__ == "__main__":
    entry_point()



from os.path import isfile
from pathlib import Path
from typing import Optional

from yaml import safe_load

from alpa.constants import CONFIG_FILE_LOCATIONS


class Config:
    def __init__(self, gh_api_token: str) -> None:
        self.gh_api_token = gh_api_token

    @staticmethod
    def _get_config_file_path() -> Optional[Path]:
        for location in CONFIG_FILE_LOCATIONS:
            expanded_path = Path(location).expanduser()
            if isfile(str(expanded_path)):
                return expanded_path

        return None

    @classmethod
    def get_config(cls, repo_name: str) -> Optional["Config"]:
        cfg_file_path = cls._get_config_file_path()
        if cfg_file_path is None:
            return None

        with open(cfg_file_path) as alpa_cfg_file:
            config_dict = safe_load(alpa_cfg_file)

        for item in config_dict["api_keys"]:
            if item["repo"]["name"] == repo_name:
                return Config(gh_api_token=item["repo"]["key"])

        return None



"""
Integration with alpa <-> upstream repo.
"""

import shutil
import subprocess
from pathlib import Path
from typing import Optional

import click
import requests
from alpa_conf import MetadataConfig
from click import ClickException
from specfile import Specfile

from alpa.repository import LocalRepo


class UpstreamIntegration(LocalRepo):
    def __init__(self, repo_path: Path) -> None:
        super().__init__(repo_path)
        self.metadata = MetadataConfig.get_config(repo_path)
        self.specfile = Specfile(Path(repo_path / f"{self.package}.spec"))
        self.name_version = (
            f"{self.specfile.expanded_name}-{self.specfile.expanded_version}"
        )
        self.nvr = f"{self.name_version}-{self.specfile.expanded_release}"

    @staticmethod
    def _find_srpm_file_from_mock_build(srpm_result_dir: Path) -> Optional[Path]:
        for file_path in srpm_result_dir.iterdir():
            if file_path.name.endswith(".src.rpm"):
                return file_path

        return None

    @staticmethod
    def _run_mock_command(cmd: list[str]) -> int:
        click.echo(f"Executing command {' '.join(cmd)}")
        return subprocess.run(cmd).returncode

    def _mock_build_one_chroot(
        self, chroot: str, result_dir: Path, source_file_name: str
    ) -> int:
        srpm_result_dir = result_dir / "srpm" / chroot
        retval = UpstreamIntegration._run_mock_command(
            [
                "mock",
                "-r",
                chroot,
                "--buildsrpm",
                "--spec",
                self.specfile.path.name,
                "--sources",
                f"{source_file_name}.tar.gz",
                "--resultdir",
                str(srpm_result_dir),
            ]
        )
        if retval != 0:
            return retval

        rpm_result_dir = result_dir / "build_results" / chroot
        srpm_path = UpstreamIntegration._find_srpm_file_from_mock_build(srpm_result_dir)
        if srpm_path is None:
            return 1

        return UpstreamIntegration._run_mock_command(
            ["mock", "-r", chroot, "--resultdir", str(rpm_result_dir), str(srpm_path)]
        )

    def _prepare_mock_result_dir(self, chroots: list[str]) -> Path:
        mock_results_path = self.repo_path / "mock_results"
        if mock_results_path.is_dir():
            # delete previous mock build
            shutil.rmtree(mock_results_path)

        for chroot in chroots:
            srpm_chroot_path = mock_results_path / "srpm" / chroot
            srpm_chroot_path.mkdir(parents=True)
            build_chroot_path = mock_results_path / "build_results" / chroot
            build_chroot_path.mkdir(parents=True)

        return mock_results_path

    @staticmethod
    def download_upstream_source(upstream_source_url: str, name_version: str) -> None:
        resp = requests.get(upstream_source_url, allow_redirects=True)
        if not resp.ok:
            raise ConnectionError(
                f"Couldn't download source from {upstream_source_url}. "
                f"Reason: {resp.reason}"
            )

        with open(f"{name_version}.tar.gz", "wb") as archive:
            archive.write(resp.content)

    def mockbuild(self, chroots: list[str]) -> None:
        with self.specfile.sources() as sources:
            source0 = min(sources, key=lambda src: src.number)

        source_file_name = source0.expanded_filename.rstrip(".tar.gz")
        self.download_upstream_source(source0.expanded_location, source_file_name)
        root_mock_result_dir = self._prepare_mock_result_dir(chroots)

        for chroot in chroots:
            retval = self._mock_build_one_chroot(
                chroot, root_mock_result_dir, source_file_name
            )
            if retval != 0:
                raise ClickException(f"Mock returned non-zero value: {retval}")



"""
These commands need to create LocalRepo -> no GH token required
"""


from os import getcwd
from pathlib import Path
from typing import List

import click
from alpa_conf import MetadataConfig
from click import ClickException, Choice
from specfile import Specfile

from alpa.config.packit import PackitConfig
from alpa.repository import LocalRepo, AlpaRepo
from alpa.upstream_integration import UpstreamIntegration

pkg_name = click.argument("name", type=str)


@click.command("show-history")
@click.option(
    "-o",
    "--oneline",
    is_flag=True,
    show_default=True,
    default=False,
    help="Each commit is displayed as one line",
)
def show_history(oneline: bool) -> None:
    """Show git history of current package"""
    params = []
    if oneline:
        params.append("--oneline")

    local_repo = LocalRepo(Path(getcwd()))
    click.echo(local_repo.get_history_of_branch(local_repo.branch, params))


@click.command("switch")
@pkg_name
def switch(name: str) -> None:
    """Switch to specified package"""
    LocalRepo(Path(getcwd())).switch_to_package(name)


@click.command("commit")
@click.option(
    "-m",
    "--message",
    type=str,
    default="",
    help="Your commit message not longer than 80 characters.",
)
@click.option("-n", "--no-verify", is_flag=True, help="Do not run pre-commit")
def commit(message: str, no_verify: bool) -> None:
    """Commit your changes in your package's repository"""
    if len(message) > 80:
        raise ClickException("Message longer than 80 characters")

    LocalRepo(Path(getcwd())).commit(message, not no_verify)


@click.command("add")
@click.argument("files", type=str, nargs=-1, required=True)
def add(files: List[str]) -> None:
    """Add files to git history. Basically calls `git add <input>`"""
    LocalRepo(Path(getcwd())).add(files)


@click.command("push")
@click.option(
    "-p",
    "--pull-request",
    is_flag=True,
    show_default=True,
    default=False,
    help="This will create pull request on GitHub for you.",
)
def push(pull_request: bool) -> None:
    """Pushes your commited changes to the Alpa repo so you can make PR"""
    repo_path = Path(getcwd())
    local_repo = LocalRepo(repo_path)

    packit_conf = PackitConfig(local_repo.package)
    if not packit_conf.packit_config_file_exists():
        packit_conf.create_packit_config()
        local_repo.git_cmd.add(".packit.yaml")
        local_repo.git_cmd.commit(
            '-m "alpa: automatically add .packit.yaml config to the package"'
        )

    local_repo.push(local_repo.branch)

    if not pull_request:
        local_repo.git_cmd.branch("-d", local_repo.feat_branch)
        return

    alpa = AlpaRepo(repo_path)
    pr = alpa.gh_repo.create_pr(
        title=f"[alpa-cli] Create update of package {local_repo.package}",
        body=(
            "This PR was created automatically via alpa-cli for "
            f"user {alpa.gh_api.gh_user}"
        ),
        source_branch=f"{local_repo.feat_branch}",
        target_branch=local_repo.package,
    )
    click.echo(f"PR#{pr.number} created. URL: {pr.html_url}")

    # go from feat branch to package branch
    local_repo.git_cmd.switch(local_repo.package)
    local_repo.git_cmd.branch("-D", local_repo.feat_branch)


@click.command("pull")
def pull() -> None:
    """Pull last recent changes of package you are on from Alpa repo"""
    local_repo = LocalRepo(Path(getcwd()))
    local_repo.pull(local_repo.branch)


@click.command("list")
@click.option("-p", "--pattern", type=str, default="", help="Optional pattern to match")
def list_(pattern: str) -> None:
    """List all packages or packages matching regex"""
    for pkg in LocalRepo(Path(getcwd())).get_packages(pattern):
        click.echo(pkg)


@click.command("genspec")
@click.option(
    "--lang",
    type=Choice(["python", "java"], case_sensitive=False),
    required=True,
    help="Choose the programming language for which the generator is designed",
)
@click.option(
    "-t",
    "--test",
    default=False,
    help=(
        "Send package with generated spec file to "
        "packit to test whether build will succeed."
    ),
)
def genspec(lang: str, test: bool) -> None:
    """This command uses some existing spec file generators for you"""
    raise NotImplementedError("Not implemented yet (1.0 goal)")


@click.command("create-packit-config")
@click.option(
    "--override",
    is_flag=True,
    show_default=True,
    default=False,
    help="Override packit config file",
)
def create_packit_config(override: bool) -> None:
    """Creates packit config based on metadata.yaml file in package"""
    if not LocalRepo(Path(getcwd())).create_packit_config(override):
        raise ClickException(
            "Packit file already exists. To override it use --override flag."
        )


def _get_chroots_to_build(meta: MetadataConfig, distros: list[str]) -> list[str]:
    chroots = []
    for arch in meta.arch:
        for distro in distros:
            chroots.append(f"{distro}-{arch}")
    return chroots


@click.command("mockbuild")
@click.option(
    "--chroot",
    type=str,
    default="",
    help="Override chroots specified in metadata.yaml by one specific chroot",
)
def mockbuild(chroot: str) -> None:
    """
    Uses mock tool to build package. WARNING: works only on rpm-based systems.

    Builds for all chroots specified in metadata.yaml. Can be overriden by --chroot
     option which does build against one specified chroot.
    """
    meta = MetadataConfig.get_config()
    distros = [chroot] if chroot else list(meta.targets)
    chroots = _get_chroots_to_build(meta, distros)
    UpstreamIntegration(Path(getcwd())).mockbuild(chroots)


@click.command("get-pkg-archive")
def get_pkg_archive() -> None:
    """Gets archive from package config"""
    specfile_path = None
    cwd = Path(getcwd())
    for file in cwd.iterdir():
        if str(file).endswith(".spec"):
            specfile_path = file

    if specfile_path is None:
        raise ClickException(f"No specfile found in {cwd}")

    specfile = Specfile(specfile_path)
    with specfile.sources() as sources:
        source0 = min(sources, key=lambda src: src.number)

    UpstreamIntegration.download_upstream_source(
        source0.expanded_location,
        f"{specfile.expanded_name}-{specfile.expanded_version}",
    )



"""
Set of commands that helps with integration of Alpa
repository.
"""
import subprocess
from os import getcwd, environ
from pathlib import Path
import re
from subprocess import call
from tempfile import NamedTemporaryFile
from typing import List, Optional, Iterable
from urllib.parse import urlparse

from click import UsageError, ClickException
import click
from git import Repo, Remote, GitCommandError

from alpa.config.packit import PackitConfig
from alpa.constants import (
    ALPA_FEAT_BRANCH,
    ALPA_FEAT_BRANCH_PREFIX,
    MAIN_BRANCH,
    ORIGIN_NAME,
    UPSTREAM_NAME,
    PackageRequest,
)
from alpa.gh import GithubAPI, GithubRepo
from alpa.messages import (
    CLONED_REPO_IS_NOT_FORK,
    NO_WRITE_ACCESS_ERR,
    NOT_IN_PREDEFINED_STATE,
    CLONED_REPO_IS_FORK,
    NO_PERMISSION_FOR_ALPA_REPO,
)


class LocalRepo:
    def __init__(self, repo_path: Path) -> None:
        self.repo_path = repo_path
        # TODO: this will differ in the future with repo_path
        self.repo_root_path = repo_path
        self.local_repo = Repo(str(self.repo_path), search_parent_directories=True)
        self.git_cmd = self.local_repo.git

        if not self._is_repo_in_predefined_state():
            raise ClickException(NOT_IN_PREDEFINED_STATE)

        # do it after predefined state is checked since this depends on it
        # (see comment in the _should_be_fork)
        self.remote_name = UPSTREAM_NAME if self._should_be_fork() else ORIGIN_NAME

        # lazy properties
        self._namespace: Optional[str] = None
        self._repo_name: Optional[str] = None

    @property
    def remote_associated_with_current_branch(self) -> str:
        return self.local_repo.active_branch.tracking_branch().remote_name

    @property
    def branch(self) -> str:
        return self.local_repo.active_branch.name

    @property
    def package(self) -> str:
        return self.branch.removeprefix(ALPA_FEAT_BRANCH_PREFIX)

    @staticmethod
    def get_feat_branch_of_package(package: str) -> str:
        return ALPA_FEAT_BRANCH.format(pkgname=package)

    @property
    def feat_branch(self) -> str:
        return ALPA_FEAT_BRANCH.format(pkgname=self.package)

    @staticmethod
    def _get_relevant_remote_refs(remote_refs: Iterable[str]) -> list[str]:
        relevant_refs = []
        for ref in remote_refs:
            parsed_ref = ref.split("/")[-1]
            if not parsed_ref.startswith(ALPA_FEAT_BRANCH_PREFIX):
                relevant_refs.append(parsed_ref)

        return relevant_refs

    def show_remote_branches(self, remote: str) -> List[str]:
        lines = [
            line.strip()
            for line in self.git_cmd.remote("--verbose", "show", remote).split("\n")
        ]
        # TODO: do a better job
        remote_branch_line = ["Remote branch:", "Remote branches:"]
        start = -1
        for line_to_match in remote_branch_line:
            if line_to_match in lines:
                start = lines.index(line_to_match)
                break

        if start == -1:
            return []

        # TODO: do a better job
        possible_start_of_local_stuff = [
            "Local branch configured for 'git pull':",
            "Local branches configured for 'git pull':",
            "Local ref configured for 'git push':",
            "Local refs configured for 'git push':",
        ]
        end = -1
        for line_to_match in possible_start_of_local_stuff:
            if line_to_match in lines:
                end = lines.index(line_to_match)
                break

        if end == -1:
            end = len(lines)

        lines_with_remote_branches = lines[start + 1 : end]
        return [line.split()[0] for line in lines_with_remote_branches]

    def get_packages(self, regex: str) -> List[str]:
        # self.local_repo.remote(name=self.remote_name).refs don't work on every case
        refs_without_main = filter(
            lambda ref: ref != "main", self.show_remote_branches(self.remote_name)
        )
        relevant_refs = self._get_relevant_remote_refs(refs_without_main)

        if regex == "":
            return list(relevant_refs)

        pattern = re.compile(regex)
        return [ref for ref in relevant_refs if pattern.match(ref)]

    def _is_repo_in_predefined_state(self) -> bool:
        remotes_name_set = {remote.name for remote in self.local_repo.remotes}
        return remotes_name_set == {ORIGIN_NAME, UPSTREAM_NAME} or remotes_name_set == {
            ORIGIN_NAME
        }

    def _should_be_fork(self) -> bool:
        remotes_name_set = {remote.name for remote in self.local_repo.remotes}
        # if repo is prepared via alpa-cli, fork should have 2 remotes and non-fork 1
        return len(remotes_name_set) > 1

    @property
    def untracked_files(self) -> list[str]:
        return self.local_repo.untracked_files

    def _get_dirty_files(self, staged: bool) -> list[str]:
        result = []
        status_output = self.git_cmd.status("--porcelain=1").split("\n")
        for line in status_output:
            file = line.split()[-1]
            if line.startswith("MM"):
                result.append(file)
                continue

            if not staged and line.startswith(" M"):
                result.append(file)
                continue

            if staged and line.startswith("M "):
                result.append(file)
                continue

        return result

    @property
    def modified_files(self) -> list[str]:
        return self._get_dirty_files(False)

    @property
    def files_to_be_committed(self) -> list[str]:
        return self._get_dirty_files(True)

    @property
    def namespace(self) -> str:
        if self._namespace is not None:
            return self._namespace

        self._namespace = self.full_reponame().split("/")[0]
        return self._namespace

    @property
    def repo_name(self) -> str:
        if self._repo_name is not None:
            return self._repo_name

        self._repo_name = self.full_reponame().split("/")[1]
        return self._repo_name

    @staticmethod
    def _format_files_to_status(files: list[str], msg: str) -> str:
        if not files:
            return ""

        output = msg + "\n"
        output += "\n".join(files)
        return output + "\n"

    def get_status_output(self) -> str:
        output = self._format_files_to_status(
            self.files_to_be_committed, "Files to commit:"
        )
        output += self._format_files_to_status(self.modified_files, "Modified files:")
        output += self._format_files_to_status(self.untracked_files, "Untracked files:")
        return output

    def branch_exists(self, branch: str) -> bool:
        for ref in self.local_repo.references:
            if ref.name == branch:
                return True

        return False

    def switch_to_package(self, package: str) -> None:
        if self.local_repo.is_dirty():
            click.echo(
                "Repo is dirty, please commit your changes before switching to"
                f" another package.\n {self.get_status_output()}"
            )
            return None

        feat_branch = self.get_feat_branch_of_package(package)
        branch_to_switch = feat_branch if self.branch_exists(feat_branch) else package
        try:
            click.echo(
                self.git_cmd.switch(branch_to_switch).replace("branch", "package")
            )
        except GitCommandError:
            # switching to the package for the first time
            click.echo(f"Switching to the package {package} for the first time")
            click.echo(self.git_cmd.fetch(self.remote_name, branch_to_switch))
            click.echo(
                self.git_cmd.switch(branch_to_switch).replace("branch", "package")
            )

    def get_history_of_branch(self, branch: str, *params: List[str]) -> str:
        return self.git_cmd.log("--decorate", "--graph", *params, branch)

    @staticmethod
    def _get_message_from_editor() -> str:
        with NamedTemporaryFile(suffix=".alpa.tmp") as temp_file:
            call([environ.get("EDITOR", "vim"), temp_file.name])
            temp_file.seek(0)
            output = temp_file.read()
            if isinstance(output, (bytes, bytearray)):
                return output.decode("utf-8")

            return output

    def _ensure_feature_branch(self) -> None:
        if self.branch != self.package:
            return None

        click.echo("Switching to feature branch")
        self.git_cmd.switch("-c", self.feat_branch)

    def commit(self, message: str, pre_commit: bool) -> bool:
        if pre_commit:
            ret = subprocess.run(["pre-commit", "run", "--all-files"])
            if ret.returncode != 0:
                return False

        self._ensure_feature_branch()
        index = self.local_repo.index
        if message:
            index.commit(message)
        else:
            index.commit(self._get_message_from_editor())

        return True

    def add(self, files: List[str]) -> None:
        self._ensure_feature_branch()
        # FIXME: alpa add . acts weird
        self.git_cmd.add(files)

    def pull(self, branch: str) -> None:
        click.echo(self.git_cmd.pull(self.remote_name, branch))

    def push(self, branch: str) -> None:
        # you always want to push to origin, even from a fork
        click.echo(self.git_cmd.push(ORIGIN_NAME, branch))

    def full_reponame(self) -> str:
        for remote in self.local_repo.remotes:
            if remote.name == self.remote_name:
                return remote.url.split(":")[-1].removesuffix(".git")

        return ""

    def create_packit_config(self, override: bool) -> bool:
        packit_conf = PackitConfig(self.package)
        if packit_conf.packit_config_file_exists() and not override:
            return False

        packit_conf.create_packit_config()
        return True

    def get_git_root(self) -> Optional[Path]:
        return Path(self.local_repo.working_tree_dir)


class AlpaRepo(LocalRepo):
    def __init__(self, repo_path: Path, gh_api: Optional[GithubAPI] = None) -> None:
        super().__init__(repo_path)

        self.gh_api = gh_api or GithubAPI(self.repo_name)
        self.gh_repo = self.gh_api.get_repo(self.namespace, self.repo_name)

    def create_package(self, package: str) -> None:
        upstream = self.gh_repo.get_upstream()
        if upstream and not upstream.has_write_access(self.gh_api.gh_user):
            raise ClickException(NO_WRITE_ACCESS_ERR)

        self.git_cmd.switch(MAIN_BRANCH)
        self.git_cmd.switch("-c", package)
        self.git_cmd.push(self.remote_name, package)
        click.echo(f"Package {package} created")

    def request_package(self, package_name: str) -> None:
        ensured_upstream = self.gh_repo.get_root_repo()
        upstream_namespace = ensured_upstream.namespace
        issue_repo = self.gh_api.get_repo(upstream_namespace, self.gh_repo.repo_name)
        issue = issue_repo.create_issue(
            PackageRequest.TITLE.value.format(package_name=package_name),
            PackageRequest.BODY.value.format(
                user=self.gh_api.gh_user,
                package_name=package_name,
                repo_name=self.gh_repo.repo_name,
            ),
        )
        issue.add_to_labels(PackageRequest.LABEL)

    def delete_package(self) -> None:
        pass

    @staticmethod
    def _prepare_cloned_repo(local_repo: Repo, gh_repo: GithubRepo) -> None:
        if not gh_repo.is_fork:
            return

        Remote.create(local_repo, UPSTREAM_NAME, gh_repo.upstream_clone_url)

    @staticmethod
    def _get_repo_name_from_url(repo_url: str) -> str:
        return repo_url.split("/")[-1].removesuffix(".git")

    @staticmethod
    def _check_for_permission_and_fork(
        clone_fork: bool, gh_repo: GithubRepo
    ) -> tuple[bool, str]:
        if clone_fork and not gh_repo.is_fork:
            return False, CLONED_REPO_IS_NOT_FORK

        if not clone_fork and gh_repo.is_fork:
            return False, CLONED_REPO_IS_FORK

        if not gh_repo.is_fork and not gh_repo.has_write_access(gh_repo.api_user):
            return False, NO_PERMISSION_FOR_ALPA_REPO

        return True, ""

    @classmethod
    def clone(cls, url: str, clone_fork: bool) -> None:
        # in case of `@` in url -> remove the `git@` prefix form it
        repo_path = urlparse(url.split("@")[-1]).path
        parsed_repo_path = repo_path.strip("/").strip(".git")
        namespace, repo_name = parsed_repo_path.split("/")
        api = GithubAPI(repo_name)
        gh_repo = api.get_repo(namespace, repo_name)

        check_result, stderr = cls._check_for_permission_and_fork(clone_fork, gh_repo)
        if not check_result:
            raise UsageError(stderr)

        cloned_repo = Repo.clone_from(
            url, f"{getcwd()}/{cls._get_repo_name_from_url(url)}"
        )
        cls._prepare_cloned_repo(cloned_repo, gh_repo)



"""
Integration with GitHub API.
"""


from os import getenv
from typing import Optional

import click
from click import UsageError
from github import Github, Issue, PullRequest, UnknownObjectException

from alpa.config.local_config import Config
from alpa.constants import GH_API_TOKEN_NAME, GH_WRITE_ACCESS
from alpa.messages import NO_GH_API_KEY_FOUND, RETURNING_CLONE_URL_MSG


class GithubRepo:
    def __init__(self, api: Github, namespace: str, repo_name: str) -> None:
        self.namespace = namespace
        self.repo_name = repo_name

        self._api = api
        self._repo = api.get_repo(f"{namespace}/{repo_name}")

    @property
    def clone_url(self) -> str:
        if self.has_write_access(self._api.get_user().login):
            return self._repo.ssh_url

        click.echo(
            RETURNING_CLONE_URL_MSG.format(
                user=self._api.get_user().login, repo=self._repo.full_name
            )
        )
        return self._repo.clone_url

    @property
    def upstream_clone_url(self) -> Optional[str]:
        upstream = self.get_upstream()
        if upstream is None:
            return None

        return upstream.clone_url

    @property
    def is_fork(self) -> bool:
        return self._repo.fork

    @property
    def api_user(self) -> str:
        return self._api.get_user().login

    def has_write_access(self, user: str) -> bool:
        for collaborator in self._repo.get_collaborators():
            if collaborator.login != user:
                continue

            return (
                self._repo.get_collaborator_permission(collaborator) in GH_WRITE_ACCESS
            )

        return False

    def get_upstream(self) -> Optional["GithubRepo"]:
        if not self.is_fork:
            return None

        namespace, repo_name = self._repo.source.full_name.split("/")
        return GithubRepo(self._api, namespace, repo_name)

    def get_root_repo(self) -> "GithubRepo":
        root_repo = self.get_upstream()
        if root_repo is None:
            root_repo = self

        return root_repo

    def create_issue(self, title: str, body: str) -> Issue:
        return self.get_root_repo()._repo.create_issue(title, body)

    def create_pr(
        self, title: str, body: str, source_branch: str, target_branch: str = "main"
    ) -> PullRequest:
        return self.get_root_repo()._repo.create_pull(
            title, body, base=target_branch, head=source_branch
        )

    def pr_exists(self, id: int) -> bool:
        try:
            self._repo.get_pull(id)
            return True
        except UnknownObjectException:
            return False


class GithubAPI:
    def __init__(self, repo_name: str) -> None:
        self._gh_api = Github(self._get_access_token(repo_name))

    @property
    def gh_user(self) -> str:
        return self._gh_api.get_user().login

    @staticmethod
    def _get_access_token(repo_name: str) -> str:
        access_token_env = getenv(GH_API_TOKEN_NAME)
        if access_token_env is not None:
            return access_token_env

        # accessing config file here since it does not make sense to do it elsewhere,
        # but in the future the config will be useful in many places
        access_token_config = Config.get_config(repo_name)
        if access_token_config is None:
            raise UsageError(NO_GH_API_KEY_FOUND.format(token=GH_API_TOKEN_NAME))

        return access_token_config.gh_api_token

    def get_repo(self, namespace: str, repo_name: str) -> GithubRepo:
        return GithubRepo(self._gh_api, namespace, repo_name)



"""
These commands need GH API token.
"""


from os import getcwd
from pathlib import Path

import click

from alpa.repository import AlpaRepo


pkg_name = click.argument("name", type=str)


@click.command("create")
@pkg_name
def create(name: str) -> None:
    """Create new package"""
    AlpaRepo(Path(getcwd())).create_package(name)


@click.command("delete")
@pkg_name
def delete(name: str) -> None:
    """Delete existing package"""
    raise NotImplementedError("Not implemented yet (1.0 goal)")


@click.command("request-package")
@pkg_name
def request_package(name: str) -> None:
    """Request new branch for new package in Alpa repo"""
    raise NotImplementedError("Not implemented yet (1.0 goal)")



from enum import Enum


UPSTREAM_NAME = "upstream"
ORIGIN_NAME = "origin"
MAIN_BRANCH = "main"


GH_API_TOKEN_NAME = "ALPA_GH_API_TOKEN"
GH_WRITE_ACCESS = ["admin", "write"]

ALPA_ISSUE_REPO_NAME = "issue-repo"
ALPA_FEAT_BRANCH_PREFIX = "__feat_"
ALPA_FEAT_BRANCH = ALPA_FEAT_BRANCH_PREFIX + "{pkgname}"

GLOBAL_CFG_FILE = "/etc/alpa.yaml"
USER_CFG_FILE = "~/.config/alpa.yaml"
# order is important! user cfg overrides global cfg
CONFIG_FILE_LOCATIONS = [USER_CFG_FILE, GLOBAL_CFG_FILE]

PACKIT_CONFIG_NAMES = [
    ".packit.yaml",
    ".packit.yml",
    ".packit.json",
    "packit.yaml",
    "packit.yml",
    "packit.json",
]


class PackageRequest(str, Enum):
    TITLE = "[alpa] New request for package {package_name}"
    BODY = "@{user} requested {package_name} for repo {repo_name}"
    LABEL = "request"



from os import getcwd
from pathlib import Path

from yaml import dump

from alpa_conf import AlpaRepoConfig, MetadataConfig

from alpa.constants import PACKIT_CONFIG_NAMES


class PackitConfig:
    def __init__(self, package_name: str) -> None:
        self.package_name = package_name
        self.working_dir = Path(getcwd())
        self.metadata = MetadataConfig.get_config(self.working_dir)
        self.alpa_repo_config = AlpaRepoConfig.get_config()

    def get_packit_config(self) -> dict:
        jobs = [
            {
                "job": "copr_build",
                "trigger": "pull_request",
                "targets": list(self.metadata.targets),
                "owner": self.alpa_repo_config.copr_owner,
                "project": f"{self.alpa_repo_config.copr_repo}-pull-requests",
            },
            {
                "job": "copr_build",
                "trigger": "commit",
                "branch": self.package_name,
                "targets": list(self.metadata.targets),
                "owner": self.alpa_repo_config.copr_owner,
                "project": self.alpa_repo_config.copr_repo,
            },
        ]

        if self.metadata.autoupdate is not None:
            autoupdate_dict = {
                "job": "copr_build",
                "trigger": "commit",
                "branch": f"__alpa_autoupdate_{self.package_name}",
                "targets": list(self.metadata.targets),
                "owner": self.alpa_repo_config.copr_owner,
                "project": f"{self.alpa_repo_config.copr_repo}-pull-requests",
            }
            jobs.append(autoupdate_dict)

        return {
            "specfile_path": f"{self.package_name}.spec",
            "srpm_build_deps": ["pip"],
            "actions": {
                "create-archive": [
                    "pip install pyalpa",
                    'bash -c "alpa get-pkg-archive"',
                    f'bash -c "ls -1 ./{self.package_name}-*.tar.gz"',
                ],
            },
            "jobs": jobs,
        }

    def packit_config_file_exists(self) -> bool:
        for packit_config_name in PACKIT_CONFIG_NAMES:
            files_in_dir = [
                file.name for file in self.working_dir.iterdir() if file.is_file()
            ]
            if packit_config_name in files_in_dir:
                return True

        return False

    def create_packit_config(self) -> None:
        if self.packit_config_file_exists():
            raise FileExistsError("Packit configuration file already exists")

        with open(".packit.yaml", "w") as packit_yaml:
            packit_yaml.write(dump(self.get_packit_config(), sort_keys=False))
